/**********************************************************  SigStream  *********************************************************

	Red Rapids
	797 North Grove Rd
	Suite 101
	Richardson, TX 75081
	www.redrapids.com | support@redrapids.com

	The following code demonstrates the operation of all Red Rapids SigStream PCI Express products.

	Note:	The application can be applied to the maximum number of Red Rapids devices (MAX_NUM_DEVICES) supported, but only
			a single device (DEV0) is demonstrated.

	Rev History:  01-10-2017:	Added support for programmable filters.
				  08-04-2016:	Added support for Model 273.
				  09-16-2015:	Added Adapter_Driver() function call to replace default driver name.
				  08-24-2015:	Zero the channel error counters after the channel status is cleared.
				  05-04-2015:	Moved the BusClose function behind all operations that access the DMA buffers.
				  01-16-2015:	Added support for Model 271.
				  08-21-2014:	Inital release.

	Copyright 2016, Red Rapids, Inc.

**********************************************************  SigStream  *********************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "RRadapter0.h"					// Adpater device driver common to all Red Rapids PCI Express products.
#include "sigstream.h"					// SigFPGA application.
#include "dma_functions.h"				// Direct Memory Access (DMA) functions.
#include "common_functions.h"			// Functions common to all Signal Stream hardware and channel types.
#include "channel_functions.h"			// Functions that apply to individual Signal Stream channels.
#include "product_functions.h"			// Functions that apply to specific Signal Stream products.
#include "filter_functions.h"			// Functions that apply to filters embedded in Signal Stream datapaths.
#include "utility_functions.h"			// Functions specific to this application example.
#include "default_operating_config.h"	// Default channel configuration values.
#include "default_hardware_config.h"	// Default device configuration values.

int main()
{
	s_Adapter *p_HW[MAX_NUM_DEVICES];			// Hardware handle to the Red Rapids device.
	s_User User;								// User defined structure passed to the interrupt service routine (ISR).
	unsigned long Status = 0;					// Variable to store status returned by function calls.
	unsigned long DMABuffer[8];					// Array to store DMA buffer allocation status for upt to eight channels (success = 0, failure = 1).
	unsigned long Performance[4] = {0,0,0,0};	// Array to store PCI RX throughput (0), TX throughput (1), RX latency (2) and TX latency (3) measurements.
	unsigned long DataBuffer[8][1024];			// Array to store data buffer allocation status for up to eight channels (success = 0, failure = 1).
	unsigned long Handle[MAX_NUM_DEVICES];		// Array to store hardware handle allocation status for each device (success = 0, failure = 1).
	unsigned long devcnt = 0;					// Variable to index device numbers in loops.
	unsigned long chancnt = 0;					// Variable to index channel numbers in loops.
	size_t pagecnt = 0;							// Variable to index page numbers in loops.
	size_t index = 0;							// Variable to index array elements in loops.
	s_Channel RXChannel[8];						// Structure to store up to eight channels of receiver configuration variables.
	s_Channel TXChannel[8];						// Structure to store up to eight channels of transmitter configuration variables.
	s_ErrorStatus ErrorStatus;					// Structure to store the results returned by a fault status query to the device.
	s_Temperature Temperature;					// Structure to store the results returned by the temperature monitor on the device.
	s_Power Power;								// Structure to store the results returned by the power monitor on the device.
	double FsDivisor = 8.0;						// Variable to set the transmitter output frequency for demonstration (Freq = Fs / FsDivisor).
	char filterfile[32];						// Array to assemble the names of filter configuration files.
	char datafile[8][12] = {"rx1data.txt", "rx2data.txt", "rx3data.txt", "rx4data.txt", "rx5data.txt", "rx6data.txt", "rx7data.txt", "rx8data.txt"};	// Files to store receiver data.
	char dmafile[8][12] = {"dma1buf.txt", "dma2buf.txt", "dma3buf.txt", "dma4buf.txt", "dma5buf.txt", "dma6buf.txt", "dma7buf.txt", "dma8buf.txt"};		// Files to store receiver dma buffer contents.
	char probefile[8][13] = {"tx1probe.txt", "tx2probe.txt", "tx3probe.txt", "tx4probe.txt", "tx5probe.txt", "tx6probe.txt", "tx7probe.txt", "tx8probe.txt"};	// Files to store 1024 DAC probe samples.

	printf("******************************** Signal Stream ********************************\n");

	//////////////////////////////////////////////  Initialize Hardware Handle  //////////////////////////////////////////////

	printf("\nInitializing structures...\n");

	// Check the DMA configuration settings for possible errors.
	for (index = 1; index <= 4; index++)
		if(DMAConfigCheck(RXDefault.DMABuffer, index) || DMAConfigCheck(TXDefault.DMABuffer, index))
		{
			// This is a fatal error since the DMA buffers cannot be correctly configured.
			switch (index)
			{
				case 1: {printf("\tERROR: Illegal channel DMA buffer BurstSizeB configuration setting.\n"); break;}
				case 2: {printf("\tERROR: Illegal channel DMA buffer BurstCount configuration setting.\n"); break;}
				case 3: {printf("\tERROR: Illegal channel DMA buffer PageCount configuration setting.\n"); break;}
				case 4: {printf("\tERROR: Illegal channel DMA buffer PagesPerMark configuration setting.\n"); break;}
			}
			Status++;
		}

	// Initialize all eight receiver and transmitter channel configuration structures even though they may not all be assigned to the hardware.
	for (chancnt = 1; chancnt <= 8; chancnt++)
	{
		// Initialize all four receiver channel configuration structures to the same default value in this example.
		RXChannel[chancnt-1] = RXDefault;
		// Initialize all four transmitter channel configuration structures to the same default value in this example.
		TXChannel[chancnt-1] = TXDefault;
	}

	// Set the driver name before attempting any communication with the hardware.
	Status += Adapter_Driver("RRadapter");

	// Initialize the hardware handle for each device.
	for (devcnt = 0; devcnt < MAX_NUM_DEVICES; devcnt++)
	{
		// Allocate memory for the structure if there were no prior errors.
		if (!Status)
		{
			p_HW[devcnt] = (s_Adapter*) malloc(sizeof(s_Adapter));
			if (p_HW[devcnt] == NULL)
			{
				// This is a fatal error because there is no hardware handle available.
				printf("\tERROR: Unable to allocate memory for the device #%lu hardware handle.\n", devcnt);
				Status++;
				// Mark this handle as a failed allocation.
				Handle[devcnt] = 1;
			}
			else
			{
				// Mark this handle as a successful allocation.
				Handle[devcnt] = 0;
				// Initialize each member of the structure.
				Adapter_Zero(p_HW[devcnt]);
				// Set a unique device number for each hardware handle.
				p_HW[devcnt]->DevNum = devcnt;
				// Set the user defined pointer to pass information to the ISR in this application.
				p_HW[devcnt]->User = &User;
			}
		}
		// Mark the handle as a failed memory allocation if there were prior errors.
		else
			Handle[devcnt] = 1;
	}

	//////////////////////////////////////////////////  Configure Hardware  //////////////////////////////////////////////////
	if (!Status)
	{
		printf("\nOpening device #%u on the host...\n", DEV0);

		// Open the Red Rapids device on the bus and only continue if successful.
		if (!BusOpen(p_HW[DEV0]))
		{

			// Print the firmware revision date to the screen.
			printf("\tFirmware revision (MMDDYYYY): %08lX\n", FirmwareRevision(p_HW[DEV0]));

			// Software reset the device.
			Status += SoftwareReset(p_HW[DEV0]);

			// Peform model number specific hardware configuration procedures.
			if (!Status)
				switch (ModelNumber(p_HW[DEV0]))
				{
					case 0x271:	// Model 271
					{
						printf("\nInitializing Red Rapids Model 271...\n");
						// Initialize hardware specific to the Model 271.
						Status += M271Initialize(p_HW[DEV0], M271Default, RXChannel);
						if (Status) printf("\tERROR: Model 271 hardware configuration failed.\n");
						// Initialize the clocks.
						Status += ClockInit(p_HW[DEV0]);
						break;
					}
					case 0x273:	// Model 273
					{
						printf("\nInitializing Red Rapids Model 273...\n");
						// Initialize hardware specific to the Model 273.
						Status += M273Initialize(p_HW[DEV0], M273Default, RXChannel);
						if (Status) printf("\tERROR: Model 273 hardware configuration failed.\n");
						// Initialize the clocks.
						Status += ClockInitLMK01000(p_HW[DEV0], M273Default.ClockDist);
						// Initiate the ADC fast start-up calibration.
						Status += AD9652Cal(p_HW[DEV0], M273_RX1TO2_ADC);
						break;
					}
					case 0x276:	// Model 276
					{
						printf("\nInitializing Red Rapids Model 276...\n");
						// Initialize hardware specific to the Model 276.
						Status += M276Initialize(p_HW[DEV0], M276Default, RXChannel);
						if (Status) printf("\tERROR: Model 276 hardware configuration failed.\n");
						// Initialize the clocks.
						Status += ClockInit(p_HW[DEV0]);
						break;
					}
					case 0x277:	// Model 277
					{
						printf("\nInitializing Red Rapids Model 277...\n");
						// Initialize hardware specific to the Model 277.
						Status += M277Initialize(p_HW[DEV0], M277Default, RXChannel);
						if (Status) printf("\tERROR: Model 277 hardware configuration failed.\n");
						// Initialize the clocks.
						Status += ClockInit(p_HW[DEV0]);
						break;
					}
					default:	// Illegal model number.
					{
						Status++;
						printf("\nERROR: Model number reported by the hardware is not recognized.\n");
						break;
					}
				}

			// Initialize the clocks if there were no prior errors.
			if (!Status)
			{
				Status = ClockInit(p_HW[DEV0]);
				// Report the clock status returned from the hardware.
				switch (ClockStatus(p_HW[DEV0], CMD_STAT1))
				{
					case 0: {printf("\tExternal sample clock active, digital clocks locked.\n"); break;}
					case 1: {printf("\tERROR: External sample clock active, digital clocks not locked.\n"); Status++; break;}
					case 2: {printf("\tSample clock locked to internal reference, digital clocks locked.\n"); break;}
					case 3: {printf("\tERROR: Sample clock locked to internal reference, digital clocks not locked.\n"); Status++; break;}
					case 4: {printf("\tSample clock locked to external reference, digital clocks locked.\n"); break;}
					case 5: {printf("\tERROR: Sample clock locked to external reference, digital clocks not locked.\n"); Status++;break;}
					default: {printf("\tERROR: Sample clock is not locked.\n"); Status++; break;}
				}
			}

			// Initialize the user defined structure attached to the hardware handle.
			for (chancnt = 1; chancnt <= User.RXChannelCount + User.TXChannelCount; chancnt++)
			{
				if (chancnt > User.RXChannelCount)
					// Initialize the structure members assigned to transmitter channels if there were no prior errors.
					if (!Status)
					{
						Status = UserInitialize(&User, chancnt, TXChannel[chancnt-1].DMABuffer, DataBuffer[chancnt-1]);
						if (Status) printf("\tERROR: Unable to allocate memory for the channel #%lu DataBuffer.\n", chancnt);
					}
					// Mark all of the buffers as a failed memory allocation if there were prior errors.
					else
						for (index = 0; index < TXChannel[chancnt-1].DMABuffer.PagesPerMark; index++)
							DataBuffer[chancnt-1][index] = 1;
				else
					// Initialize the structure members assigned to receiver channels if there were no prior errors.
					if (!Status)
					{
						Status = UserInitialize(&User, chancnt, RXChannel[chancnt-1].DMABuffer, DataBuffer[chancnt-1]);
						if (Status) printf("\tERROR: Unable to allocate memory for the channel #%lu DataBuffer.\n", chancnt);
					}
					// Mark all of the buffers as a failed memory allocation if there were prior errors.
					else
						for (index = 0; index < RXChannel[chancnt-1].DMABuffer.PagesPerMark; index++)
							DataBuffer[chancnt-1][index] = 1;
			}
				
			// Set hardware configuration options if there were no prior errors.
			if (!Status)
			{
				// Select the sources that will illuminate LED-A and LED-B.
				Status = LED(p_HW[DEV0], LEDSourceDefault);
				// Set the time of day (TOD) consisting of the seconds value and the optional clock frequency.
				LoadTODSettings(p_HW[DEV0], TODSettingsDefault);
				// Print the current seconds value.
				printf("\tCurrent time of day (TOD) setting: %lu seconds\n", GetTODSeconds(p_HW[DEV0]));
				// Set the interrupt mask bits for individual fault conditions.
				Status = ErrorMask(p_HW[DEV0], ErrorMaskDefault);
				// Set the interrupt mask bits for individual channel status events.
				for (chancnt = 1; chancnt <= User.RXChannelCount + User.TXChannelCount; chancnt++)
					Status += ChannelMask(p_HW[DEV0], chancnt, ChannelMaskDefault);
				// Set the global interrupt mask.
				Status += InterruptMask(p_HW[DEV0], GlobalMaskDefault);
				// Report any errors.
				if (Status) printf("\tERROR: Unable to set hardware configuration options.\n");
			}

			// Configure each channel with application specific settings.
			for (chancnt = 1; chancnt <= User.RXChannelCount + User.TXChannelCount; chancnt++)
				// Configure the channel if there were no prior errors.
				if (!Status)
				{
					if (chancnt <= User.RXChannelCount)
					{
						// Load receiver channel configuration settings.
						Status = LoadChannel(p_HW[DEV0], chancnt, RXChannel[chancnt-1]);
						// Determine whether there are programmable filters to configure.
						// Note: Programmable filters became available after 2016 firmware releases.
						if (((FirmwareRevision(p_HW[DEV0]) & 0x000000000000FFFFull) > 0x0000000000002016ull) & (RXChannel[chancnt - 1].DatapathControl.Transform[0] > 1))
							// Configure all of the available filters.
							for (index = 1; index <= GetFilterCount(p_HW[DEV0], chancnt); index++)
							{
								// Generate a filename of the form filter#_type$.txt, where # is the filter number and $ is the type.
								FilterFilename(p_HW[DEV0], chancnt, index, filterfile);
								// Configure each filter using the parameters stored in filename filter#_type$.txt under the "filters" subdirectory.
								FilterConfigure(p_HW[DEV0], chancnt, index, filterfile);
							}

					}
					
					else
						// Load transmitter channel configuration settings.
						Status = LoadChannel(p_HW[DEV0], chancnt, TXChannel[chancnt-1]);
					// A status value of 100 or greater indicates a failure to allocate the requested DMA buffer.
					if (Status < 100)
					{
						// Mark the successful allocation of a DMA buffer for this channel.
						DMABuffer[chancnt-1] = 0;
						// Transmitter channels must be preloaded with data before the initial DMA transfer is requested.
						if (chancnt > User.RXChannelCount)
						{
							// Load DAC samples into the TX data buffer, a sine wave is used here for demonstration purposes (Frequency = Fs / FsDivisor).
							// The TX data buffer will refresh the DMA buffer as data is consumed, but the contents remain static in this example.
							if (DataLoad(User.p_DataBuffer[chancnt-1], TXChannel[chancnt-1], FsDivisor, 16))
								printf("\tINFO: Discontinuities will be present in the channel #%lu TX waveform.\n", chancnt);
							// Preoload the DMA buffer with samples stored in the TX data buffer.
							for(pagecnt=0; pagecnt < TXChannel[chancnt-1].DMABuffer.PageCount; pagecnt++)
								memcpy(p_HW[DEV0]->DMA.VirtDMAAdr[chancnt-1][pagecnt], User.p_DataBuffer[chancnt-1][pagecnt%TXChannel[chancnt-1].DMABuffer.PagesPerMark], p_HW[DEV0]->DMA.BurstSizeB[chancnt-1]*p_HW[DEV0]->DMA.BurstCount[chancnt-1]);
						}
					}
					else
					{
						// Mark the failed allocation of a DMA buffer for this channel.
						DMABuffer[chancnt-1] = 1;
						printf("\tERROR: Unable to allocate memory for the channel #%lu DMABuffer.\n", chancnt);
					}
				}
				// Mark the DMA buffer as a failed memory allocation if there were prior errors.
				else
					DMABuffer[chancnt-1] = 1;

			///////////////////////////////////////////////  Process Channels  ///////////////////////////////////////////////

			// Start channel processing if there were no prior errors and the receiver to transmitter loopback is not selected.
			if (!Status && TXDefault.SwitchSelect.ChannelSelect != 7)
			{
				// Measure the maximum achievable PCI throughput by initiating continuous DMA transactions on all channels.
				// Note: This is not required to run an application, but it provides valuable system performance information.
				printf("\nMeasuring the PCI performance...\n");
				// Collect PCI performance data with maximum DMA transfers on all of the channels operating simultaneously.
				PciBenchmark(p_HW[DEV0]);
				// Retrieve the measured performance.
				PciPerformance(p_HW[DEV0], Performance);
				// Report the measured instantaneous PCI throughput.
				if (User.RXChannelCount) printf("\tMeasured RX PCI performance: %.3f Gbytes/sec\n", Performance[0]/1e9);
				if (User.TXChannelCount) printf("\tMeasured TX PCI performance: %.3f Gbytes/sec\n", Performance[1]/1e9);
				// Report the measured maximum PCI latency recorded.
				if (User.RXChannelCount) printf("\tMeasured RX PCI latency: %.3f usec\n", Performance[2]*4e-3);
				if (User.TXChannelCount) printf("\tMeasured TX PCI latency: %.3f nsec\n", Performance[3]*4e-3);
				// Clear the PCI latency performance measurement.
				ClearLatency(p_HW[DEV0]);

				// Start processing data on all available channels simultaneously.
				printf("\nStarting channels...\n");
				// Enable each channel to start processing when the specified start event occurs.
				for (chancnt = 1; chancnt <= User.RXChannelCount + User.TXChannelCount; chancnt++)
				{
					// Clear the channel status register and error counters.
					Status = ClearChannelStatus(p_HW[DEV0], chancnt);
					User.Error[chancnt-1][0] = 0;
					User.Error[chancnt-1][1] = 0;
					User.Error[chancnt-1][2] = 0;
					// Enable the channel if there were no configuration errors.
					FlowControl(p_HW[DEV0], chancnt, 1);
					printf("\tChannel #%lu started.\n", chancnt);
				}

				// Monitor the status of all channels with a polling loop to determine when they are ready for a start trigger.
				do
				{
					Status = 0;
					// Pause before the next status query to minimize PCI bus traffic.
					MicrosecondTimer(100);
					// Check the ready status of each channel (0 = Not Ready, 1 = Ready).
					for (chancnt = 1; chancnt <= User.RXChannelCount + User.TXChannelCount; chancnt++)
						Status += ReadyStatus(p_HW[DEV0], chancnt);
				}
				while (Status != User.RXChannelCount + User.TXChannelCount);

				// Display an update message that includes the interrupt status.
				if (GlobalMaskDefault)
					printf("\nWaiting for all channels to stop processing (interrupts enabled)...\n");
				else
					printf("\nWaiting for all channels to stop processing (interrupts disabled)...\n");
				// It is not safe to abort execution while DMA transactions are active.
				printf("\tWARNING: CTRL-C FROM THE KEYBOARD MAY CRASH THE COMPUTER!\n");

				// Issue a software trigger to begin processing data on all channels.
				// Note: Different start/stop triggers can be selected in the channel configuration.
				// Note: Begin monitoring channel status immediately following the trigger so that DMA data is not lost.
				SoftwareTrigger(p_HW[DEV0]);
				
				// Monitor the status of all channels with a polling loop and terminate when all channels stop processing.
				do
				{
					Status = 0;
					// Check the on/off status of each channel to determine when they have completed processing.
					for (chancnt = 1; chancnt <= User.RXChannelCount + User.TXChannelCount; chancnt++)
						Status += FlowStatus(p_HW[DEV0], chancnt);
					// Process all channels for any status changes if interrupts are not enabled.
					if (!GlobalMaskDefault)
						ProcessChannels(p_HW[DEV0]);
					// Pause before the next channel status query to minimize PCI bus traffic.
					// Note: The pause should be much less than the time interval between DMA markers.
					if (Status) MicrosecondTimer(100);
				}
				while (Status);

				//////////////////////////////////////////////  Report Status  //////////////////////////////////////////////

				printf("\nRetrieving temperature status...\n");
				// Report the values measured by the temperature monitor.
				Status = GetTemperature(p_HW[DEV0], &Temperature);
				if (!Status)
				{
					printf("\tINFO: FPGA die temperature is %.2f degrees Celsius.\n", Temperature.Sensor[0]);
					printf("\tINFO: ADC proximity temperature is %.2f degrees Celsius.\n", Temperature.Sensor[1]);
					printf("\tINFO: QDR proximity temperature is %.2f degrees Celsius.\n", Temperature.Sensor[2]);
					printf("\tINFO: Shielded proximity temperature is %.2f degrees Celsius.\n", Temperature.Sensor[3]);
					printf("\tINFO: Card center proximity temperature is %.2f degrees Celsius.\n", Temperature.Sensor[4]);
				}
				else
					printf("\tERROR: Temperature monitor returned an error.\n");

				printf("\nRetrieving power status...\n");
				// Report the values measured by the power monitor.
				printf("\tINFO: Total power dissipation is %.2f Watts.\n", GetPower(p_HW[DEV0], &Power));
				printf("\tINFO: Supply #1 summary: %.2f Volts, %.2f Amps, %.2f Watts\n", Power.Supply1[0], Power.Supply1[1], Power.Supply1[2]);
				printf("\tINFO: Supply #2 summary: %.2f Volts, %.2f Amps, %.2f Watts\n", Power.Supply2[0], Power.Supply2[1], Power.Supply2[2]);
				// The third power supply is not present on native PCI Express products, do not report if the measured current is zero.
				if (Power.Supply3[1]) printf("\tINFO: Supply #3 summary: %.2f Volts, %.2f Amps, %.2f Watts\n", Power.Supply3[0], Power.Supply3[1], Power.Supply3[2]);

				printf("\nReviewing channel status...\n");
				// Report any faults that were detected during execution.
				Status = GetErrorStatus(p_HW[DEV0], &ErrorStatus);
				// Print each fault detected.
				if (ErrorStatus.Bar[0]) printf("\tERROR: Write to an illegal BAR0 address detected.\n");
				if (ErrorStatus.Bar[1]) printf("\tERROR: Read from an illegal BAR0 address detected.\n");
				if (ErrorStatus.Bar[2]) printf("\tERROR: Write to an illegal BAR2 address detected.\n");
				if (ErrorStatus.Bar[3]) printf("\tERROR: Read from an illegal BAR2 address detected.\n");
				if (ErrorStatus.Clock[0]) printf("\tERROR: IDELAY out of lock condition detected.\n");
				if (ErrorStatus.Clock[1]) printf("\tERROR: Digital clock initialization failed to complete.\n");
				if (ErrorStatus.Clock[2]) printf("\tERROR: QDR-A SRAM calibration failed.\n");
				if (ErrorStatus.Clock[3]) printf("\tERROR: QDR-B SRAM calibration failed.\n");

				// Report any channel errors that were detected during execution.
				for (chancnt = 1; chancnt <= User.RXChannelCount + User.TXChannelCount; chancnt++)
				{
					if (User.Error[chancnt-1][0]) printf("\tERROR: Channel #%lu converter (ADC/DAC) error detected.\n", chancnt);
					if (User.Error[chancnt-1][1]) printf("\tERROR: Channel #%lu FIFO overflow/underflow, check PCI performance.\n", chancnt);
					if (User.Error[chancnt-1][2]) printf("\tERROR: Channel #%lu DMA data was lost, adjust the DMA marker to correct.\n", chancnt);
				}

				// Report the interrupt response time if interrupts were enabled.
				if (GlobalMaskDefault) printf("\tINFO: The last interrupt was serviced in %.2f microseconds.\n", 0.004*InterruptResponse(p_HW[DEV0]));

				// Report any data residue remaining in receiver DMA FIFOs.
				for (chancnt = 1; chancnt <= User.RXChannelCount; chancnt++)
					if (FIFOResidue(p_HW[DEV0], chancnt)) printf("\tINFO: Data residue in the channel #%lu DMA FIFO: %lu bytes\n", chancnt, FIFOResidue(p_HW[DEV0], chancnt));

				// Extract the contents of each TX probe and save to a file.
				if (User.TXChannelCount)
				{
					printf("\nSaving probe data on all transmitter channels...\n");
					// Record the first 1024 samples that were sent to the DAC input.
					for (chancnt = User.RXChannelCount+1; chancnt <= User.RXChannelCount + User.TXChannelCount; chancnt++)
					{
						printf("\tWriting the first 1024 DAC samples on channel #%lu to file %s.\n", chancnt, probefile[chancnt-1]);
						ProbeSave(p_HW[DEV0], chancnt, probefile[chancnt-1]);
					}
				}

				// Extract the contents of each RX buffer and save to a file.
				if (User.RXChannelCount)
				{
					printf("\nSaving captured data on all receiver channels...\n");
					for (chancnt = 1; chancnt <= User.RXChannelCount; chancnt++)
						// Verify that data was packetized if payload packing was enabled, otherwise there is no way to find the first sample in the buffer.
						if (RXChannel[chancnt-1].PayloadFormat.PayloadPacking && !RXChannel[chancnt-1].IFdpktFormat.PacketSize)
							printf("\tERROR: Cannot extract packed data on channel #%lu without packets.\n", chancnt);
						else
						{
							// Save the first page of each RX data buffer to a file.
							printf("\tWriting channel #%lu data buffer contents to file %s.\n", chancnt, datafile[chancnt-1]);
							Status = DataSave(User.p_DataBuffer[chancnt-1], RXChannel[chancnt-1], TODSettingsDefault, datafile[chancnt-1], 1, 0);
							// Save the first page of each DMA buffer to a file.
							printf("\tWriting channel #%lu DMA buffer contents to file %s.\n", chancnt, dmafile[chancnt-1]);
							Status = DataSave(p_HW[DEV0]->DMA.VirtDMAAdr[chancnt-1], RXChannel[chancnt-1], TODSettingsDefault, dmafile[chancnt-1], 1, 1);
						}
				}
				printf("\n...all operations completed.\n");

			}
			else
				if ((TXDefault.SwitchSelect.ChannelSelect == 7) && (Status < 100))
					printf("\nThe RX to TX loopback channel configuration is selected...\n");
				else
					printf("\nERROR: Terminating execution due to hardware configuration errors.\n");

			// Free the memory allocated to the DMA buffers.
			for (chancnt = 1; chancnt <= User.RXChannelCount + User.TXChannelCount; chancnt++)
				if (!DMABuffer[chancnt-1]) Status = DMAFree(p_HW[DEV0], chancnt);

			// Close the Red Rapids device on the bus.
			Status = BusClose(p_HW[DEV0]);
			
		}
		else
			printf("\n...terminating execution because device could not be opened.\n");
	}
	else
		printf("\n...terminating execution due to hardware handle intialization errors.\n");


	/////////////////////////////////////////////////  Free Allocated Memory /////////////////////////////////////////////////

	// Free the memory allocated to the RX data buffers.
	for (chancnt = 1; chancnt <= User.RXChannelCount; chancnt++)
		for (index = 0; index < RXChannel[chancnt-1].DMABuffer.PagesPerMark; index++)
			if (!DataBuffer[chancnt-1][index]) free(User.p_DataBuffer[chancnt-1][index]);

	// Free the memory allocated to the TX data buffers.
	for (chancnt = User.RXChannelCount + 1; chancnt <= User.RXChannelCount + User.TXChannelCount; chancnt++)
		for (index = 0; index < TXChannel[chancnt-1].DMABuffer.PagesPerMark; index++)
			if (!DataBuffer[chancnt-1][index]) free(User.p_DataBuffer[chancnt-1][index]);

	// Free the memory allocated to the hardware handles.
	for (devcnt = 0; devcnt < MAX_NUM_DEVICES; devcnt++)
		if (!Handle[devcnt]) free(p_HW[devcnt]);

return 0;
}
