/****************************************************  Channel Functions  *****************************************************

	Red Rapids
	797 North Grove Rd
	Suite 101
	Richardson, TX 75081
	www.redrapids.com | support@redrapids.com

	The following functions are specific to individual Signal Stream channels.

	Rev History:  06-08-2016:	Changed a structure definition to support transceiver products.
				  08-21-2014:	Inital release.

	Copyright 2014, Red Rapids, Inc.

****************************************************  Channel Functions  *****************************************************/

#include <stdlib.h>

#include "RRadapter0.h"
#include "sigstream.h"
#include "dma_functions.h"
#include "common_functions.h"
#include "channel_functions.h"

// Set the channel interrupt mask bits to enable (1) or disable (0) interrupt generation for each event type.
// Return of zero indicates success, non-zero indicates an error condition.
unsigned long ChannelMask(s_Adapter *p_HW, unsigned long ChannelNumber, s_ChannelIntMask ChannelIntMask)
{
	return(Write(p_HW, BAR2, ChannelNumber << 12 | CHANNEL_INT_MASK,
				(unsigned long long)ChannelIntMask.DMAMarker |
				(unsigned long long)ChannelIntMask.Error[0] << 4 |
				(unsigned long long)ChannelIntMask.Error[1] << 5));
}

// Issue a reset to restore the channel configuration registers to their power-on state.
// Return of zero indicates success, non-zero indicates an error condition.
unsigned long ChannelReset(s_Adapter *p_HW, unsigned long ChannelNumber)
{
	return(Write(p_HW, BAR2, ChannelNumber << 12 | CHANNEL_RESET, 0x0ull));
}

// Clear the channel status register.
// Return of zero indicates success, non-zero indicates an error condition.
unsigned long ClearChannelStatus(s_Adapter *p_HW, unsigned long ChannelNumber)
{
	// Clear the channel status register.
	return(Write(p_HW, BAR2, ChannelNumber << 12 | CHANNEL_STATUS, 0x0ull));
}

// Start (OnOff = 1) or stop (OnOff = 0) channel processing.
// Return of zero indicates success, non-zero indicates an error condition.
unsigned long FlowControl(s_Adapter *p_HW, unsigned long ChannelNumber, unsigned long OnOff)
{
	return(Write(p_HW, BAR2, ChannelNumber << 12 | SYNCHRONIZER_CTL, (unsigned long long)OnOff));
}

// Report the current on/off status of the requested channel.
// Return of zero indicates the channel is off, one indicates the channel is on.
unsigned long FlowStatus(s_Adapter *p_HW, unsigned long ChannelNumber)
{
	unsigned long long Value = 0;
	
	Read(p_HW, BAR2, ChannelNumber << 12 | SYNCHRONIZER_CTL, &Value);
	
	return((unsigned long)(Value & 0x0000000000000001ull));
}

// Report the channel status and clear the values.
// Returns the total number of status events detected.
unsigned long GetChannelStatus(s_Adapter *p_HW, unsigned long ChannelNumber, s_ChannelStatus *p_ChannelStatus)
{
	unsigned long long Value = 0;
	unsigned long Status = 0;
	size_t cnt = 0;
	
	// Save and clear the contents of the channel status register.
	Write(p_HW, BAR2, ChannelNumber << 12 | CHANNEL_STATUS, 0x0ull);
	// Read the channel status register.
	Read(p_HW, BAR2, ChannelNumber << 12 | CHANNEL_STATUS, &Value);

	// Strip the DMA marker bit.
	if ((p_ChannelStatus->DMAMarker = Value & 0x0000000000000001ull)) Status++;

	// Strip the individual channel error indicator bits.
	for (cnt = 0; cnt < 2; cnt++)
		if ((p_ChannelStatus->Error[cnt] = Value >> ((cnt + 4)) & 0x0000000000000001ull)) Status++;

	// Return the total fault count.
	return(Status);
}

// Set all of the channel configuration registers.
// Return of zero indicates success, non-zero indicates an error condition.
unsigned long LoadChannel(s_Adapter *p_HW, unsigned long ChannelNumber, s_Channel Channel)
{
	unsigned long Status = 0;
	
		// Allocate DMA buffers for the selected channel.
		Status = 100*DMAAllocate(p_HW, ChannelNumber, Channel.DMABuffer);
		
		// Load the SWITCH SELECT configuration register for the selected channel.
		Status += LoadSwitchSelect(p_HW, ChannelNumber, Channel.SwitchSelect);
		
		// Load the PAYLOAD FORMAT configuration register for the selected channel.
		Status += LoadPayloadFormat(p_HW, ChannelNumber, Channel.PayloadFormat);
		
		// Load the DATAPATH CONTROL configuration register for the selected channel.
		Status += LoadDatapathControl(p_HW, ChannelNumber, Channel.DatapathControl);
		
		// Load the SYNCHRONIZER COMMAND configuration register for the selected channel.
		Status += LoadSynchronizer(p_HW, ChannelNumber, Channel.Synchronizer);
		
		// Load the TOD TRIGGERS configuration register for the selected channel.
		Status += LoadTODTriggers(p_HW, ChannelNumber, Channel.TODTriggers);
		
		// Load the EVENT DURATION configuration register for the selected channel.
		Status += LoadEventDuration(p_HW, ChannelNumber, Channel.EventDuration);
		
		// Load the IF DATA PACKET FORMAT configuration register for the selected channel.
		Status += LoadIFdpktFormat(p_HW, ChannelNumber, Channel.IFdpktFormat);

	return(Status);
}

// Configure the datapath operating controls.
// Return of zero indicates success, non-zero indicates an error condition.
unsigned long LoadDatapathControl(s_Adapter *p_HW, unsigned long ChannelNumber, s_DatapathControl DatapathControl)
{
	return(Write(p_HW, BAR2, ChannelNumber << 12 | DATAPATH_CONTROL,
			(unsigned long long)DatapathControl.Transform[1] |
			(unsigned long long)DatapathControl.Residue[0] << 24 |
			(unsigned long long)DatapathControl.Residue[1] << 28 |
			(unsigned long long)DatapathControl.SyncInitMask[0] << 32 |
			(unsigned long long)DatapathControl.SyncInitMask[1] << 33 |
			(unsigned long long)DatapathControl.SyncInitMask[2] << 34 |
			(unsigned long long)DatapathControl.SyncInitMask[3] << 35 |
			(unsigned long long)DatapathControl.SyncCycleMask[0] << 48 |
			(unsigned long long)DatapathControl.SyncCycleMask[1] << 49 |
			(unsigned long long)DatapathControl.SyncCycleMask[2] << 50 |
			(unsigned long long)DatapathControl.SyncCycleMask[3] << 51)
			+
			Write(p_HW, BAR2, ChannelNumber << 12 | TUNER_CONTROL,
				(-1*(unsigned long long)DatapathControl.Transform[2] & 0x00000000007FFFFFull) |
				((-1*(unsigned long long)DatapathControl.Transform[3] << 32) & 0x007FFFFF00000000ull))
			+
			Write(p_HW, BAR2, ChannelNumber << 12 | FILTER_CONTROL,
				(unsigned long long)DatapathControl.Transform[0]));
}

// Set the optional event duration timers.
// Return of zero indicates success, non-zero indicates an error condition.
unsigned long LoadEventDuration(s_Adapter *p_HW, unsigned long ChannelNumber, s_EventDuration EventDuration)
{
	return(Write(p_HW, BAR2, ChannelNumber << 12 | EVENT_DURATION1,
				(unsigned long long)EventDuration.SampleCount |
				(unsigned long long)EventDuration.BurstCount << 32)
			+
			Write(p_HW, BAR2, ChannelNumber << 12 | EVENT_DURATION2,
				(unsigned long long)EventDuration.CycleCount |
				(unsigned long long)EventDuration.FrameCount << 32));
}

// Configure the IF data packet to the requested format.
// Return of zero indicates success, non-zero indicates an error condition.
unsigned long LoadIFdpktFormat(s_Adapter *p_HW, unsigned long ChannelNumber, s_IFdpktFormat IFdpktFormat)
{
	return(Write(p_HW, BAR2, ChannelNumber << 12 | IF_DPKT_FORMAT1,
				(unsigned long long)IFdpktFormat.PacketSize |
				(unsigned long long)IFdpktFormat.TSFSelect << 16 |
				(unsigned long long)IFdpktFormat.TSISelect << 17 |
				(unsigned long long)IFdpktFormat.TrailerSelect << 18 |
				(unsigned long long)IFdpktFormat.ClassIDSelect << 19 |
				(unsigned long long)IFdpktFormat.StreamIDSelect << 20 |
				(unsigned long long)IFdpktFormat.StreamID << 32)
			+
			Write(p_HW, BAR2, ChannelNumber << 12 | IF_DPKT_FORMAT2,
				(unsigned long long)IFdpktFormat.ClassID)
			+
			Write(p_HW, BAR2, ChannelNumber << 12 | IF_DPKT_FORMAT3,
				(unsigned long long)IFdpktFormat.UserIndicators[0] |
				(unsigned long long)IFdpktFormat.UserIndicators[1] << 1 |
				(unsigned long long)IFdpktFormat.UserIndicators[2] << 2 |
				(unsigned long long)IFdpktFormat.SpectralInversion << 4 |
				(unsigned long long)IFdpktFormat.ContextPktCntEnable << 7 |
				(unsigned long long)IFdpktFormat.IndicatorEnables << 8));
}

// Configure the IF data payload to the requested format.
// Return of zero indicates success, non-zero indicates an error condition.
unsigned long LoadPayloadFormat(s_Adapter *p_HW, unsigned long ChannelNumber, s_PayloadFormat PayloadFormat)
{
	return(Write(p_HW, BAR2, ChannelNumber << 12 | PAYLOAD_FORMAT,
			(unsigned long long)PayloadFormat.PayloadPacking |
			(unsigned long long)PayloadFormat.Endianness << 2 |
			(unsigned long long)PayloadFormat.DataItemSize << 4 |
			(unsigned long long)PayloadFormat.EventTagOrder << 12 |
			(unsigned long long)PayloadFormat.EventTagMask << 16 |
			(unsigned long long)PayloadFormat.EventTagIndex << 20 |
			(unsigned long long)PayloadFormat.ItemPackingFieldSize << 24 |
			(unsigned long long)PayloadFormat.ChannelTag << 32));
}

// Select the requested data source for the channel.
// Return of zero indicates success, non-zero indicates an error condition.
unsigned long LoadSwitchSelect(s_Adapter *p_HW, unsigned long ChannelNumber, s_SwitchSelect SwitchSelect)
{
	return(Write(p_HW, BAR2, ChannelNumber << 12 | SWITCH_SELECT,
			(unsigned long long)SwitchSelect.ChannelSelect |
			(unsigned long long)SwitchSelect.Amplitude << 16));
}

// Configure the synchronizer operating controls.
// Return of zero indicates success, non-zero indicates an error condition.
unsigned long LoadSynchronizer(s_Adapter *p_HW, unsigned long ChannelNumber, s_Synchronizer Synchronizer)
{
	return(Write(p_HW, BAR2, ChannelNumber << 12 | SYNCHRONIZER_CMD,
			(unsigned long long)Synchronizer.StartEvent |
			(unsigned long long)Synchronizer.StartOffset << 4 |
			(unsigned long long)Synchronizer.StopEvent << 20 |
			(unsigned long long)Synchronizer.StopOffset << 24 |
			(unsigned long long)Synchronizer.CycleSelect << 40 |
			(unsigned long long)Synchronizer.CycleStop << 44));
}

// Set the optional time of day (TOD) start and stop triggers.
// Return of zero indicates success, non-zero indicates an error condition.
unsigned long LoadTODTriggers(s_Adapter *p_HW, unsigned long ChannelNumber, s_TODTriggers TODTriggers)
{
	return(Write(p_HW, BAR2, ChannelNumber << 12 | TOD_TRIGGERS1,
				(unsigned long long)TODTriggers.FractionalSecondsStart |
				(unsigned long long)TODTriggers.IntegerSecondsStart << 32)
			+
			Write(p_HW, BAR2, ChannelNumber << 12 | TOD_TRIGGERS2,
				(unsigned long long)TODTriggers.FractionalSecondsStop |
				(unsigned long long)TODTriggers.IntegerSecondsStop << 32));
}

// Report the current ready status of the requested channel.
// Return of zero indicates the channel is not ready for a start trigger, one indicates the channel is ready.
unsigned long ReadyStatus(s_Adapter *p_HW, unsigned long ChannelNumber)
{
	unsigned long long Value = 0;
	
	Read(p_HW, BAR2, ChannelNumber << 12 | SYNCHRONIZER_CTL, &Value);
	
	// Shift the ready status bit so that the function returns either one or zero.
	return((unsigned long)((Value >> 1) & 0x0000000000000001ull));
}
